// Copyright (C) 2008-2009 - INRIA - Michael Baudin
// Copyright (C) 2010 - DIGITEO - Michael Baudin
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

function [flag,errmsg] = assert_almostequal ( varargin )
  //  Check that computed and expected are numerically close.
  //
  // Calling Sequence
  //   flag = assert_almostequal ( computed , expected , precision )
  //   flag = assert_almostequal ( computed , expected , precision , comptype )
  //   [flag,errmsg] = assert_almostequal ( ... )
  //
  // Parameters
  //   computed: a 1x1 matrix of doubles, the computed result
  //   expected : a 1x1 matrix of doubles, the expected result
  //   precision : a 1x1 matrix of doubles, the relative or absolute tolerance
  //   comptype : a 1x1 matrix of strings, "matrix" or "element" (default comptype="matrix"). The comparison type.
  //   flag : a 1x1 matrix of boolean, %t if computed is close to expected, %f if not
  //   errmsg : a 1x1 matrix of strings, the error message. If flag==%t, then errormsg=="". If flag==%f, then errmsg contains the error message.
  //
  // Description
  //   Performs silently if the two matrices of doubles computed and expected are close.
  //
  //   If expected is nonzero, then the tolerance is considered as 
  //   relative.
  //   If expected is zero, then the tolerance is considered as absolute.
  //
  //   The default expected value should be nonzero, so that the 
  //   comparison is based on a relative error.
  //   This allows to assert the number of significant digits in the computed result.
  //
  //   This procedure only works when the computed and expected variables 
  //   are matrices of doubles. 
  //   It will generate an error in any other case.
  //
  //   If the comparison shows that computed is not almost equal to expected, 
  //   <itemizedlist>
  //   <listitem>if the errmsg output variable is not used, an error is generated,</listitem>
  //   <listitem>if the errmsg output variable is used, no error is generated.</listitem>
  //   </itemizedlist>
  //
  //   In the process of comparing the values, we separate %nan, +%inf, -%inf and remaining values.
  //   Comparing nan values between them is not possible. This is why we compare the 
  //   indices where %nan value occurs.
  //   If we form differences of infinity values, we produce %nan values.
  //   This is why we compare the indices where +%inf values occurs. 
  //   We do the same for -%inf values.
  //   Then, the non-nan, non-infinity values are actually compared.
  //
  //   The default value comptype="matrix" option performs the comparison for the matrices as a whole, 
  //   the norm of the difference of the matrices is used.
  //   The comptype="element" option performs the comparison elementwise, i.e. 
  //   all the elements of the matrices must be almost equal.
  //
  //   The computed and expected values are not exchangable, i.e. the 
  //   order of the input arguments matters.
  //   Indeed, the comparison is based on the value of the expected 
  //   variable. Hence, using the computed value instead of the expected 
  //   value may lead to two sorts of unexpected results. 
  //   First, a test might pass instead of failing. Second, a test 
  //   might fail instead of passing. See below for examples of this.
  //
  //   When the expected value is nonzero, the precision should be 
  //   set to a multiple of the machine precision %eps.
  //   The precision should also be chosen with the lowest 
  //   possible magnitude, that is, we shoud configure 
  //   the precision as accurately as possible.
  //   During the process of configuring the precision, we 
  //   suggest to use the following precisions, in that order :
  //   %eps (all digits correct), 10*%eps, 100*%eps, 1.e3*%eps, 1.e4*%eps, ..., 
  //   1.e17*%eps (no digit correct).
  //   See below for examples of this.
  //
  //   In general, the computed variable is the result of a computation, 
  //   while expected should be a constant, such as 12.345 or 
  //   1.23456e7, for example. 
  //   Indeed, this allows to make sure of
  //   the type of tolerance used, that is, it allows to 
  //   predict with certainty whether a relative or 
  //   absolute tolerance is to be used.
  //
  // Examples
  //   // Comparisons which are successful.
  //   // Relative error :
  //   assert_almostequal ( 1.0 , 1.0 , %eps ); 
  //   assert_almostequal ( ones(10,1) , ones(10,1) , %eps );
  //   // Absolute error :
  //   assert_almostequal ( 1.23456789123456789e-30 , 0 , 1.e-10 ); 
  //   assert_almostequal ( [1.0 %nan], [1.0 %nan] , %eps );
  //
  //   // Comparisons which are failures.
  //   // Error message produced :
  //   assert_almostequal ( 1.0 , 2.0 , %eps ); 
  //   // Error message produced :
  //   flag = assert_almostequal ( 1.0 , 2.0 , %eps )
  //   // No error message produced :
  //   [flag,errmsg] = assert_almostequal ( 1.0 , 2.0 , %eps )
  //   assert_almostequal ( 1 , [2 3] , %eps );
  //   assert_almostequal ( [%nan 1.0], [1.0 %nan] , %eps );
  //   assert_almostequal ( 1.0 + 5 * %eps , 1.0 , %eps );
  //   assert_almostequal ( 1.23456789123456789e-30 , 1.3e-30 , %eps );
  //
  //   // In the case where expected is nonzero, the 
  //   // tolerance for relative error should be a 
  //   // multiple of %eps.
  //   // The following test is a success and shows 
  //   // that less than 11 digits are lost with respect 
  //   // to the maximum possible accuracy.
  //   assert_almostequal ( 1.23456 , 1.23457 , 1.e11*%eps );
  //
  //   // We must not exchange computed and expected.
  //   // The following test pass: we use an absolute tolerance
  //   // because the expected value is zero.
  //   assert_almostequal ( 1.23456789e-30 , 0 , 1.e-10 );
  //   // The following test fail: we use a relative tolerance
  //   // because the expected value is nonzero.
  //   assert_almostequal ( 0 , 1.23456789e-30 , 1.e-10 );
  //
  //   // We must configure the precision as tightly as possible.
  //   // The following test fails, because the precision is too small
  //   // with respect to the given numbers.
  //   assert_almostequal ( 1.23456 , 1.23457 , %eps );
  //   // In order to get the number of lost digits, 
  //   // we can use the following formula :
  //   ceil(log10(norm(computed-expected)/norm(expected)/%eps))
  //   // which returns 11.
  //   // We now make a judgment on the accuracy and conclude 
  //   // that 11 digits lost is acceptable: the following test pass.
  //   assert_almostequal ( 1.23456 , 1.23457 , 1.e11*%eps );
  //
  //   // We give here a practical example of the use of assert_almostequal.
  //   // We solve an averagely ill-conditionned system of linear
  //   // equations, based on the Hilbert matrix.
  //   n = 6;
  //   // The expected value is set by ourselves.
  //   expected = ones(n,1);
  //   A = testmatrix("hilb",n);
  //   // We compute the condition number of the matrix : ~10^8
  //   ceil(log10(cond(A)))
  //   // This means that the number of digits lost,  
  //   // predicted by theory, is 8.
  //   // The right-hand side is computed given A and expected.
  //   b = A * expected;
  //   // In this case, a Gauss algorithm with partial 
  //   // pivoting is used.
  //   computed = A\b;
  //   // The following test fails: we have lost some digits.
  //   assert_almostequal(computed,expected,%eps)
  //   // We compute the actual number of lost digits: 5
  //   ceil(log10(norm(computed-expected)/norm(expected)/%eps))
  //   // We accept this computation, since the actual number of lost 
  //   // digits (i.e. 5) is close - but lower - than what was  
  //   // predicted by theory (i.e. 8).
  //   // The following test pass.
  //   assert_almostequal(computed,expected,1.e5*%eps);
  //
  //   // The following test pass.
  //   assert_almostequal ( [1 1.e5] , [2 1.e5] , 1.e-3 )
  //   // We see that 1 is not almost equal to 2: we must use the elementwise comparison.
  //   // The following test does not pass.
  //   assert_almostequal ( [1 1.e5] , [2 1.e5] , 1.e-3 , "element" )
  //
  //   // The following test pass.
  //   // It is non-trivial to implement it.
  //   [flag,errmsg] = assert_almostequal ( [1.2345 %inf -%inf %nan] , [1.2346 %inf -%inf %nan] , 1.e-4 )
  //
  // Authors
  //   Michael Baudin, Digiteo, 2009-2010

  [lhs,rhs]=argn()
  if ( and(rhs <> [3 4] ) ) then
    errmsg = sprintf ( gettext ( "%s: Unexpected number of input arguments : %d provided while %d to %d are expected.") , "assert_almostequal" , rhs , 3 , 4 )
    error(errmsg)
  end
  //
  // Get arguments
  computed = varargin(1)
  expected = varargin(2)
  precision = varargin(3)
  if ( rhs < 4 ) then
    comptype = "matrix"
  else
    comptype = varargin(4)
  end
  //
  // Check types of variables
  if ( typeof(computed) <> "constant" ) then
    errmsg = sprintf ( gettext ( "%s: Unexpected type of input argument #%d : variable %s has type %s while %s is expected.") , "assert_almostequal" , 1 , "computed" , typeof(computed) , "constant" )
    error(errmsg)
  end
  if ( typeof(expected) <> "constant" ) then
    errmsg = sprintf ( gettext ( "%s: Unexpected type of input argument #%d : variable %s has type %s while %s is expected.") , "assert_almostequal" , 2 , "expected" , typeof(expected) , "constant" )
    error(errmsg)
  end
  if ( typeof(precision) <> "constant" ) then
    errmsg = sprintf ( gettext ( "%s: Unexpected type of input argument #%d : variable %s has type %s while %s is expected.") , "assert_almostequal" , 3 , "precision" , typeof(precision) , "constant" )
    error(errmsg)
  end  
  if ( typeof(comptype) <> "string" ) then
    errmsg = sprintf ( gettext ( "%s: Unexpected type of input argument #%d : variable %s has type %s while %s is expected.") , "assert_almostequal" , 4 , "comptype" , typeof(comptype) , "string" )
    error(errmsg)
  end  
  //
  // Check sizes of variables
  if ( size(precision,"*") <> 1 ) then
    errmsg = sprintf ( gettext ( "%s: Unexpected size of input argument #%d : variable %s has size %d while %d is expected.") , "assert_almostequal" , 3 , "precision" , size(precision,"*") , 1 )
    error(errmsg)
  end
  if ( size(comptype,"*") <> 1 ) then
    errmsg = sprintf ( gettext ( "%s: Unexpected size of input argument #%d : variable %s has size %d while %d is expected.") , "assert_almostequal" , 4 , "comptype" , size(comptype,"*") , 1 )
    error(errmsg)
  end
  //
  // Check values of variables
  if ( and ( comptype <> ["matrix" "element"] ) ) then
    errmsg = sprintf ( gettext ( "%s: Unexpected value input argument #%d : variable %s has value %s.") , "assert_almostequal" , 4 , "comptype" , comptype )
    error(errmsg)
  end
  //
  // Proceed...
  ncom = size(computed)
  nexp = size(expected)
  if ( or(ncom <> nexp) ) then
    cstr = "[" + strcat(string(ncom)," ") + "]"
    estr = "[" + strcat(string(nexp)," ") + "]"
    errmsg = sprintf ( gettext ( "%s: We have size(computed) = %s while size(expected) = %s: not consistent comparison.") , "assert_almostequal" , cstr , estr )
    error(errmsg)
  end  
  //
  // Decompose the expected value into nan indices, inf indices and regular indices
  // This allows to solve the following issue:
  // if computed is %inf and expected is %inf, the difference is %nan,
  // which makes the computations fail.	
  [kcpinf , kcninf , kcnan , kcreg , creg] = infnanindices ( computed )
  [kepinf , keninf , kenan , kereg , ereg] = infnanindices ( expected )
  //
  if ( comptype == "matrix" ) then
    if ( norm(ereg) == 0 ) then
      areclose = ( norm ( creg - ereg ) <= precision )
    else
      areclose = ( norm ( creg - ereg ) <= precision * norm(ereg) )
    end
  else
    // Create a matrix of true values
    nt = size(ereg,"*")
    entries = (zeros(nt,1) == zeros(nt,1))
    // Set the comparison for absolute tolerance entries
    kz = find(ereg==0)
    compm = ( abs ( creg(kz) - ereg(kz) ) <= precision )
    // Use the (:) trick to make a column vector out of the matrix of booleans
    entries(kz) = compm(:)
    // Set the comparison for relative tolerance entries
    knz = find(ereg<>0)
    compm = ( abs ( creg(knz) - ereg(knz) ) <= precision * abs(ereg(knz)) )
    entries(knz) = compm(:)
    // Compute the global condition from the entries conditions
    areclose = and(entries)
  end
  //
  // The regular values must be almost equal and 
  // * the +%inf must be at the same place,
  // * the -%inf must be at the same place,
  // * the %nan must be at the same place.
  if ( areclose & and(kcpinf == kepinf) & and(kcninf == keninf) & and(kcnan == kenan) ) then
    flag = %t
    errmsg = ""
  else
    flag = %f
    // Change the message if the matrix contains more than one value
    if ( size(expected,"*") == 1 ) then
      estr = string(expected)
    else
      estr = "[" + string(expected(1)) + " ...]"
    end
    if ( size(computed,"*") == 1 ) then
      cstr = string(computed)
    else
      cstr = "[" + string(computed(1)) + " ...]"
    end
    pstr = string(precision)
    errmsg = msprintf(gettext("%s: Assertion failed: expected = %s while computed = %s with precision=%s"), ..
      "assert_almostequal",estr,cstr,pstr)
    if ( lhs < 2 ) then
      // If no output variable is given, generate an error
      error ( errmsg )
    end
  end
endfunction

// Returns the indices where 
//  * kpinf : x == +%inf, 
//  * kninf : x == -%inf, 
//  * knan : x is a %nan 
//  * kreg : x is not an infinity, not a nan
//  * xreg = x(kreg)
// These 4 sets of indices have no intersection.
//
// Example :
// x = [1 2 3 -%inf %inf %nan %inf %nan -%inf 4 5 6]
// [kpinf , kninf , knan , kreg , xreg] = infnanindices ( x )
// xreg = [1 2 3 4 5 6]
// kreg = [1 2 3 10 11 12]
// knan = [6 8]
// kninf = [4 9]
// kpinf = [5 7]
function [kpinf , kninf , knan , kreg , xreg] = infnanindices ( x )
  kpinf = find(x==%inf)
  kninf = find(x==-%inf)
  knan = find(isnan(x))
  kreg = find(abs(x)<>%inf & ~isnan(x))
  xreg = x(kreg)
endfunction



